/*
 * Copyright 2023-2024 FalsePattern
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.falsepattern.zigbrains.lsp.editor;

import com.falsepattern.zigbrains.common.util.ApplicationUtil;
import com.falsepattern.zigbrains.common.util.FileUtil;
import com.falsepattern.zigbrains.lsp.actions.LSPReferencesAction;
import com.falsepattern.zigbrains.lsp.client.languageserver.ServerOptions;
import com.falsepattern.zigbrains.lsp.client.languageserver.requestmanager.RequestManager;
import com.falsepattern.zigbrains.lsp.client.languageserver.wrapper.LanguageServerWrapper;
import com.falsepattern.zigbrains.lsp.contributors.fixes.LSPCodeActionFix;
import com.falsepattern.zigbrains.lsp.contributors.fixes.LSPCommandFix;
import com.falsepattern.zigbrains.lsp.contributors.icon.LSPIconProvider;
import com.falsepattern.zigbrains.lsp.contributors.psi.LSPPsiElement;
import com.falsepattern.zigbrains.lsp.contributors.rename.LSPRenameProcessor;
import com.falsepattern.zigbrains.lsp.listeners.LSPCaretListenerImpl;
import com.falsepattern.zigbrains.lsp.requests.Timeout;
import com.falsepattern.zigbrains.lsp.requests.Timeouts;
import com.falsepattern.zigbrains.lsp.requests.WorkspaceEditHandler;
import com.falsepattern.zigbrains.lsp.utils.DocumentUtils;
import com.falsepattern.zigbrains.lsp.utils.FileUtils;
import com.falsepattern.zigbrains.lsp.utils.GUIUtils;
import com.intellij.codeInsight.completion.InsertionContext;
import com.intellij.codeInsight.daemon.DaemonCodeAnalyzer;
import com.intellij.codeInsight.lookup.AutoCompletionPolicy;
import com.intellij.codeInsight.lookup.LookupElement;
import com.intellij.codeInsight.lookup.LookupElementBuilder;
import com.intellij.codeInsight.template.TemplateManager;
import com.intellij.codeInsight.template.impl.TemplateImpl;
import com.intellij.codeInsight.template.impl.TextExpression;
import com.intellij.lang.annotation.Annotation;
import com.intellij.lang.annotation.AnnotationHolder;
import com.intellij.lang.annotation.HighlightSeverity;
import com.intellij.openapi.actionSystem.ActionManager;
import com.intellij.openapi.command.CommandProcessor;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.EditorModificationUtil;
import com.intellij.openapi.editor.LogicalPosition;
import com.intellij.openapi.editor.ScrollType;
import com.intellij.openapi.editor.SelectionModel;
import com.intellij.openapi.editor.event.DocumentEvent;
import com.intellij.openapi.editor.event.DocumentListener;
import com.intellij.openapi.editor.event.EditorMouseEvent;
import com.intellij.openapi.editor.ex.EditorSettingsExternalizable;
import com.intellij.openapi.fileEditor.FileDocumentManager;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.fileEditor.OpenFileDescriptor;
import com.intellij.openapi.fileTypes.PlainTextLanguage;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.TextRange;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiFile;
import com.intellij.ui.Hint;
import com.intellij.util.SmartList;
import lombok.val;
import org.apache.commons.lang3.StringUtils;
import org.eclipse.lsp4j.CodeAction;
import org.eclipse.lsp4j.CodeActionContext;
import org.eclipse.lsp4j.CodeActionParams;
import org.eclipse.lsp4j.Command;
import org.eclipse.lsp4j.CompletionItem;
import org.eclipse.lsp4j.CompletionItemKind;
import org.eclipse.lsp4j.CompletionList;
import org.eclipse.lsp4j.CompletionParams;
import org.eclipse.lsp4j.DeclarationParams;
import org.eclipse.lsp4j.DefinitionParams;
import org.eclipse.lsp4j.Diagnostic;
import org.eclipse.lsp4j.DidSaveTextDocumentParams;
import org.eclipse.lsp4j.DocumentFormattingParams;
import org.eclipse.lsp4j.DocumentRangeFormattingParams;
import org.eclipse.lsp4j.ExecuteCommandParams;
import org.eclipse.lsp4j.FormattingOptions;
import org.eclipse.lsp4j.InlayHint;
import org.eclipse.lsp4j.InlayHintParams;
import org.eclipse.lsp4j.InsertReplaceEdit;
import org.eclipse.lsp4j.InsertTextFormat;
import org.eclipse.lsp4j.Location;
import org.eclipse.lsp4j.LocationLink;
import org.eclipse.lsp4j.Position;
import org.eclipse.lsp4j.Range;
import org.eclipse.lsp4j.ReferenceContext;
import org.eclipse.lsp4j.ReferenceParams;
import org.eclipse.lsp4j.RenameParams;
import org.eclipse.lsp4j.SignatureHelp;
import org.eclipse.lsp4j.SignatureHelpParams;
import org.eclipse.lsp4j.SignatureInformation;
import org.eclipse.lsp4j.TextDocumentIdentifier;
import org.eclipse.lsp4j.TextDocumentSaveReason;
import org.eclipse.lsp4j.TextDocumentSyncKind;
import org.eclipse.lsp4j.TextEdit;
import org.eclipse.lsp4j.WillSaveTextDocumentParams;
import org.eclipse.lsp4j.WorkspaceEdit;
import org.eclipse.lsp4j.jsonrpc.JsonRpcException;
import org.eclipse.lsp4j.jsonrpc.messages.Either;
import org.eclipse.lsp4j.jsonrpc.messages.Tuple;
import org.eclipse.lsp4j.services.TextDocumentService;
import org.jetbrains.annotations.NotNull;

import javax.swing.Icon;
import java.awt.Point;
import java.lang.ref.WeakReference;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.WeakHashMap;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.BiFunction;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.falsepattern.zigbrains.common.util.ApplicationUtil.computableReadAction;
import static com.falsepattern.zigbrains.common.util.ApplicationUtil.computableWriteAction;
import static com.falsepattern.zigbrains.common.util.ApplicationUtil.invokeLater;
import static com.falsepattern.zigbrains.common.util.ApplicationUtil.pool;
import static com.falsepattern.zigbrains.common.util.ApplicationUtil.writeAction;
import static com.falsepattern.zigbrains.lsp.utils.DocumentUtils.toEither;

/**
 * Class handling events related to an Editor (a Document)
 * <p>
 * editor              The "watched" editor
 * mouseListener       A listener for mouse clicks
 * mouseMotionListener A listener for mouse movement
 * documentListener    A listener for keystrokes
 * selectionListener   A listener for selection changes in the editor
 * wrapper.getRequestManager()      The related wrapper.getRequestManager(), connected to the right LanguageServer
 * serverOptions       The options of the server regarding completion, signatureHelp, syncKind, etc
 * wrapper             The corresponding LanguageServerWrapper
 */
public class EditorEventManager {

    public final DocumentEventManager documentEventManager;
    protected static Logger LOG = Logger.getInstance(EditorEventManager.class);

    public Editor editor;
    public LanguageServerWrapper wrapper;
    private Project project;
    private TextDocumentIdentifier identifier;
    private LSPCaretListenerImpl caretListener;

    public List<String> completionTriggers;
    private List<String> signatureTriggers;
    private TextDocumentSyncKind syncKind;
    private volatile boolean needSave = false;
    private long predTime = -1L;
    private long ctrlTime = -1L;
    private boolean isOpen = false;

    private boolean mouseInEditor = true;
    private Hint currentHint;

    private final List<Diagnostic> diagnostics = new ArrayList<>();
    private AnnotationHolder anonHolder;
    private List<Annotation> annotations = new ArrayList<>();
    private WeakHashMap<Annotation, List<QuickFixRequest>> quickFixes = new WeakHashMap<>();
    private volatile boolean diagnosticSyncRequired = true;
    private volatile boolean codeActionSyncRequired = false;

    private static final long CTRL_THRESH = EditorSettingsExternalizable.getInstance().getTooltipsDelay() * 1000000;

    public static final String SNIPPET_PLACEHOLDER_REGEX = "(\\$\\{\\d+:?([^{^}]*)}|\\$\\d+)";

    //Todo - Revisit arguments order and add remaining listeners
    public EditorEventManager(Editor editor, DocumentListener documentListener, LSPCaretListenerImpl caretListener,
                              RequestManager requestmanager, ServerOptions serverOptions, LanguageServerWrapper wrapper) {

        this.editor = editor;
        this.wrapper = wrapper;
        this.caretListener = caretListener;

        this.identifier = new TextDocumentIdentifier(FileUtils.editorToURIString(editor));
        this.syncKind = serverOptions.syncKind;

        this.completionTriggers = (serverOptions.completionOptions != null
                && serverOptions.completionOptions.getTriggerCharacters() != null) ?
                serverOptions.completionOptions.getTriggerCharacters() :
                new ArrayList<>();

        this.signatureTriggers = (serverOptions.signatureHelpOptions != null
                && serverOptions.signatureHelpOptions.getTriggerCharacters() != null) ?
                serverOptions.signatureHelpOptions.getTriggerCharacters() :
                new ArrayList<>();

        this.project = editor.getProject();

        EditorEventManagerBase.registerManager(this);

        this.currentHint = null;

        this.documentEventManager = new DocumentEventManager(editor.getDocument(), documentListener, syncKind, wrapper);
    }

    @SuppressWarnings("unused")
    public Project getProject() {
        return project;
    }

    @SuppressWarnings("unused")
    public RequestManager getRequestManager() {
        return wrapper.getRequestManager();
    }

    @SuppressWarnings("unused")
    public TextDocumentIdentifier getIdentifier() {
        return identifier;
    }

    private boolean isSupportedLanguageFile(PsiFile file) {
        return file.getLanguage().isKindOf(PlainTextLanguage.INSTANCE)
                || FileUtils.isFileSupported(file.getVirtualFile());
    }

    /**
     * Returns the position of the definition given a position in the editor
     *
     * @param position The position
     * @return The location of the definition
     */
    private Location requestDeclaration(Position position) {
        return requestDefinitionOrDeclaration(position, Timeouts.DECLARATION, DeclarationParams::new, TextDocumentService::declaration);
    }

    /**
     * Returns the position of the declaration given a position in the editor
     *
     * @param position The position
     * @return The location of the declaration
     */
    private Location requestDefinition(Position position) {
        return requestDefinitionOrDeclaration(position, Timeouts.DEFINITION, DefinitionParams::new, TextDocumentService::definition);
    }

    private <T> Location requestDefinitionOrDeclaration(Position position, Timeouts timeout, BiFunction<TextDocumentIdentifier, Position, T> paramsConstructor, BiFunction<TextDocumentService, T, CompletableFuture<Either<List<? extends Location>, List<? extends LocationLink>>>> requester) {
        val params = paramsConstructor.apply(identifier, position);
        val request = requester.apply(wrapper.getRequestManager(), params);

        if (request == null) {
            return null;
        }
        try {
            val result = request.get(Timeout.getTimeout(timeout), TimeUnit.MILLISECONDS);
            wrapper.notifySuccess(timeout);
            if (result == null) {
                return null;
            }
            if (result.isLeft() && !result.getLeft().isEmpty()) {
                return result.getLeft().get(0);
            } else if (result.isRight() && !result.getRight().isEmpty()) {
                var def = result.getRight().get(0);
                return new Location(def.getTargetUri(), def.getTargetRange());
            }
        } catch (TimeoutException e) {
            LOG.warn(e);
            wrapper.notifyFailure(timeout);
            return null;
        } catch (InterruptedException | JsonRpcException | ExecutionException e) {
            LOG.warn(e);
            wrapper.crashed(e);
            return null;
        }
        return null;
    }

    public Pair<List<PsiElement>, List<VirtualFile>> references(int offset) {
        return references(offset, false, false);
    }

    public Pair<WeakReference<PsiElement>, Pair<List<PsiElement>, List<VirtualFile>>> renameCache = null;

    public Pair<List<PsiElement>, List<VirtualFile>> referencesForRename(PsiElement originalElement, boolean getOriginalElement, boolean close) {
        if (renameCache != null) {
            var cached = renameCache.getFirst().get();
            if (cached == originalElement) {
                return renameCache.getSecond();
            }
        }
        var newValue = references(originalElement.getTextOffset(), getOriginalElement, close);
        renameCache = new Pair<>(new WeakReference<>(originalElement), newValue);
        return newValue;
    }

    public Pair<List<PsiElement>, List<VirtualFile>> referencesForRename(int offset, boolean getOriginalElement, boolean close) {
        if (renameCache != null) {
            var cached = renameCache.getFirst().get();
            if (cached != null) {
                var textRange = cached.getTextRange();
                if (textRange.contains(offset) || textRange.contains(offset-1)) {
                    return renameCache.getSecond();
                }
            }
        }
        renameCache = null;
        return references(offset, getOriginalElement, close);
    }

    /**
     * Returns the references given the position of the word to search for
     * Must be called from main thread
     *
     * @param offset The offset in the editor
     * @return An array of PsiElement
     */
    public Pair<List<PsiElement>, List<VirtualFile>> references(int offset, boolean getOriginalElement, boolean fast) {
        renameCache = null;
        Position lspPos = DocumentUtils.offsetToLSPPos(editor, offset);
        TextDocumentIdentifier textDocumentIdentifier = new TextDocumentIdentifier(FileUtils.editorToURIString(editor));
        ReferenceParams params = new ReferenceParams(textDocumentIdentifier, lspPos, new ReferenceContext(getOriginalElement));
        params.setPosition(lspPos);
        params.setTextDocument(identifier);
        CompletableFuture<List<? extends Location>> request = wrapper.getRequestManager().references(params);
        if (request != null) {
            try {
                List<? extends Location> res = request.get(Timeout.getTimeout(Timeouts.REFERENCES), TimeUnit.MILLISECONDS);
                wrapper.notifySuccess(Timeouts.REFERENCES);
                if (res != null && !res.isEmpty()) {
                    List<VirtualFile> openedEditors = fast ? null : new ArrayList<>();
                    List<PsiElement> elements = new ArrayList<>();
                    res.forEach(l -> {
                        Position start = l.getRange().getStart();
                        Position end = l.getRange().getEnd();
                        String uri = FileUtil.sanitizeURI(l.getUri());
                        VirtualFile file = FileUtil.virtualFileFromURI(uri);
                        if (fast) {
                            if (file == null)
                                return;
                            val document = FileDocumentManager.getInstance().getDocument(file);
                            if (document == null)
                                return;
                            val psiFile = PsiDocumentManager.getInstance(project).getPsiFile(document);
                            if (psiFile == null)
                                return;
                            val element = psiFile.findElementAt(DocumentUtils.LSPPosToOffset(document, start));
                            if (element == null)
                                return;
                            elements.add(element);
                        } else {
                            Editor curEditor = FileUtils.editorFromUri(uri, project);
                            if (curEditor == null && file != null) {
                                OpenFileDescriptor descriptor =
                                        new OpenFileDescriptor(project, file, start.getLine(), start.getCharacter());
                                curEditor = computableWriteAction(
                                        () -> FileEditorManager.getInstance(project).openTextEditor(descriptor, false));
                                openedEditors.add(file);
                            }
                            if (curEditor == null) {
                                LOG.warn("Error occurred in LSP references.");
                                return;
                            }
                            int logicalStart = DocumentUtils.LSPPosToOffset(curEditor, start);
                            int logicalEnd = DocumentUtils.LSPPosToOffset(curEditor, end);
                            String name = curEditor.getDocument().getText(new TextRange(logicalStart, logicalEnd));
                            elements.add(new LSPPsiElement(name, project, logicalStart, logicalEnd,
                                                           PsiDocumentManager.getInstance(project)
                                                                             .getPsiFile(curEditor.getDocument())));
                        }
                    });
                    return new Pair<>(elements, openedEditors);
                } else {
                    return new Pair<>(null, null);
                }
            } catch (TimeoutException e) {
                LOG.warn(e);
                wrapper.notifyFailure(Timeouts.REFERENCES);
                return new Pair<>(null, null);
            } catch (InterruptedException | JsonRpcException | ExecutionException e) {
                LOG.warn(e);
                wrapper.crashed(e);
                return new Pair<>(null, null);
            }
        }
        return new Pair<>(null, null);
    }

    /**
     * @return The current diagnostics highlights
     */
    public synchronized List<Diagnostic> getDiagnostics() {
        this.diagnosticSyncRequired = false;
        return this.diagnostics;
    }

    /**
     * @return The current diagnostic annotations
     */
    public synchronized List<Annotation> getAnnotations() {
        this.codeActionSyncRequired = false;
        return this.annotations;
    }

    public synchronized Map<Annotation, List<QuickFixRequest>> fetchQuickFixes() {
        return quickFixes;
    }

    public synchronized void setAnnotations(List<Annotation> annotations) {
        this.annotations = annotations;
    }

    public synchronized void setAnonHolder(AnnotationHolder holder) {
        this.anonHolder = holder;
    }

    public synchronized boolean isDiagnosticSyncRequired() {
        return this.diagnosticSyncRequired;
    }

    public synchronized boolean isCodeActionSyncRequired() {
        return this.codeActionSyncRequired;
    }

    /**
     * Applies the diagnostics to the document
     *
     * @param diagnostics The diagnostics to apply from the server
     */
    public void diagnostics(List<Diagnostic> diagnostics) {

        // If both of the old diagnostics and the received diagnostics are empty, we can simply return without
        // re-triggering the annotator.
        if (editor.isDisposed() || (this.diagnostics.isEmpty() && diagnostics.isEmpty())) {
            return;
        }

        synchronized (this.diagnostics) {
            this.diagnostics.clear();
            this.diagnostics.addAll(diagnostics);
            diagnosticSyncRequired = true;
            // Triggers force full DaemonCodeAnalyzer execution.
            updateErrorAnnotations();
        }
    }

    /**
     * Retrieves the commands needed to apply a CodeAction
     *
     * @param offset The cursor position(offset) which should be evaluated for code action request.
     * @return The list of commands, or null if none are given / the request times out
     */
    @SuppressWarnings("WeakerAccess")
    public List<Either<Command, CodeAction>> codeAction(int offset) {
        CodeActionParams params = new CodeActionParams();
        params.setTextDocument(identifier);
        Range range = new Range(DocumentUtils.offsetToLSPPos(editor, offset),
                DocumentUtils.offsetToLSPPos(editor, offset));
        params.setRange(range);

        // Calculates the diagnostic context.
        List<Diagnostic> diagnosticContext = new ArrayList<>();
        synchronized (this.diagnostics) {
            diagnostics.forEach(diagnostic -> {
                int startOffset = DocumentUtils.LSPPosToOffset(editor, diagnostic.getRange().getStart());
                int endOffset = DocumentUtils.LSPPosToOffset(editor, diagnostic.getRange().getEnd());
                if (offset >= startOffset && offset <= endOffset) {
                    diagnosticContext.add(diagnostic);
                }
            });
        }

        CodeActionContext context = new CodeActionContext(diagnosticContext);
        params.setContext(context);
        CompletableFuture<List<Either<Command, CodeAction>>> future = wrapper.getRequestManager().codeAction(params);
        if (future != null) {
            try {
                List<Either<Command, CodeAction>> res = future.get(Timeout.getTimeout(Timeouts.CODEACTION), TimeUnit.MILLISECONDS);
                wrapper.notifySuccess(Timeouts.CODEACTION);
                return res;
            } catch (TimeoutException e) {
                LOG.warn(e);
                wrapper.notifyFailure(Timeouts.CODEACTION);
                return null;
            } catch (InterruptedException | JsonRpcException | ExecutionException e) {
                LOG.warn(e);
                wrapper.crashed(e);
                return null;
            }
        }
        return null;
    }

    public CompletableFuture<SignatureHelp> getSignatureHelp(int offset) {
        if (editor.isDisposed()) {
            return CompletableFuture.failedFuture(new RuntimeException("Editor is disposed"));
        }
        SignatureHelpParams params = new SignatureHelpParams(identifier, DocumentUtils.offsetToLSPPos(editor, offset));
        val result = new CompletableFuture<SignatureHelp>();
        pool(() -> {
            val future = wrapper.getRequestManager().signatureHelp(params);
            if (future == null) {
                result.complete(null);
                return;
            }
            try {
                val signatureResp = future.get(Timeout.getTimeout(Timeouts.SIGNATURE), TimeUnit.MILLISECONDS);
                wrapper.notifySuccess(Timeouts.SIGNATURE);
                if (signatureResp == null) {
                    result.complete(null);
                    return;
                }
                result.complete(signatureResp);
            } catch (TimeoutException e) {
                LOG.warn(e);
                wrapper.notifyFailure(Timeouts.SIGNATURE);
                result.completeExceptionally(e);
            } catch (JsonRpcException | ExecutionException | InterruptedException e) {
                LOG.warn(e);
                wrapper.crashed(e);
                result.completeExceptionally(e);
            } catch (Exception e) {
                LOG.warn("Internal error occurred when processing signature help");
                result.completeExceptionally(e);
            }
        });
        return result;
    }

    private String extractLabel(SignatureInformation signatureInformation, Either<String, Tuple.Two<Integer, Integer>> label) {
        if (label.isLeft()) {
            return label.getLeft();
        } else if (label.isRight()) {
            return signatureInformation.getLabel().substring(label.getRight().getFirst(), label.getRight().getSecond());
        } else {
            return "";
        }
    }

    /**
     * Reformat the whole document
     */
    public void reformat() {
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            DocumentFormattingParams params = new DocumentFormattingParams();
            params.setTextDocument(identifier);
            FormattingOptions options = new FormattingOptions();
            options.setTabSize(DocumentUtils.getTabSize(editor));
            options.setInsertSpaces(DocumentUtils.shouldUseSpaces(editor));
            params.setOptions(options);

            CompletableFuture<List<? extends TextEdit>> request = wrapper.getRequestManager().formatting(params);
            if (request == null) {
                return;
            }
            try {
                request.thenAccept(formatting -> {
                    if (formatting != null) {
                        invokeLater(() -> applyEdit(toEither(formatting), "Reformat document", false));
                    }
                });
            } catch (IndexOutOfBoundsException e) {
                //Race condition, user will inevitably retry. Just log.
                LOG.error("Reformatting race condition", e);
            }
        });
    }

    /**
     * Reformat the text currently selected in the editor
     */
    public void reformatSelection() {
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            DocumentRangeFormattingParams params = new DocumentRangeFormattingParams();
            params.setTextDocument(identifier);
            SelectionModel selectionModel = editor.getSelectionModel();
            int start = computableReadAction(selectionModel::getSelectionStart);
            int end = computableReadAction(selectionModel::getSelectionEnd);
            Position startingPos = DocumentUtils.offsetToLSPPos(editor, start);
            Position endPos = DocumentUtils.offsetToLSPPos(editor, end);
            params.setRange(new Range(startingPos, endPos));
            // Todo - Make Formatting Options configurable
            FormattingOptions options = new FormattingOptions();
            options.setTabSize(DocumentUtils.getTabSize(editor));
            options.setInsertSpaces(DocumentUtils.shouldUseSpaces(editor));
            params.setOptions(options);

            CompletableFuture<List<? extends TextEdit>> request = wrapper.getRequestManager().rangeFormatting(params);
            if (request == null) {
                return;
            }
            request.thenAccept(formatting -> {
                if (formatting == null) {
                    return;
                }
                invokeLater(() -> {
                    if (!editor.isDisposed()) {
                        applyEdit(toEither(formatting), "Reformat selection", false);
                    }
                });
            });
        });
    }

    public void rename(String renameTo) {
        rename(renameTo, editor.getCaretModel().getCurrentCaret().getOffset());
    }

    /**
     * Rename a symbol in the document
     *
     * @param renameTo The new name
     */
    public void rename(String renameTo, int offset) {
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            Position servPos = DocumentUtils.offsetToLSPPos(editor, offset);
            RenameParams params = new RenameParams(identifier, servPos, renameTo);
            CompletableFuture<WorkspaceEdit> request = wrapper.getRequestManager().rename(params);
            if (request != null) {
                request.thenAccept(res -> {
                    WorkspaceEditHandler
                            .applyEdit(res, "Rename to " + renameTo, new ArrayList<>(LSPRenameProcessor.getEditors()));
                    LSPRenameProcessor.clearEditors();
                });
            }
        });
    }

    /**
     * Returns the completion suggestions given a position
     *
     * @param pos The LSP position
     * @return The suggestions
     */
    public Iterable<? extends LookupElement> completion(Position pos) {

        List<LookupElement> lookupItems = new ArrayList<>();
        CompletableFuture<Either<List<CompletionItem>, CompletionList>> request = wrapper.getRequestManager()
                .completion(new CompletionParams(identifier, pos));
        if (request == null) {
            return lookupItems;
        }

        try {
            Either<List<CompletionItem>, CompletionList> res = request.get(Timeout.getTimeout(Timeouts.COMPLETION), TimeUnit.MILLISECONDS);
            wrapper.notifySuccess(Timeouts.COMPLETION);
            if (res == null) {
                return lookupItems;
            }
            if (res.getLeft() != null) {
                for (CompletionItem item : res.getLeft()) {
                    LookupElement lookupElement = createLookupItem(item);
                    if (lookupElement != null) {
                        lookupItems.add(lookupElement);
                    }
                }
            } else if (res.getRight() != null) {
                for (CompletionItem item : res.getRight().getItems()) {
                    LookupElement lookupElement = createLookupItem(item);
                    if (lookupElement != null) {
                        lookupItems.add(lookupElement);
                    }
                }
            }
        } catch (TimeoutException | InterruptedException e) {
            LOG.warn(e);
            wrapper.notifyFailure(Timeouts.COMPLETION);
        } catch (JsonRpcException | ExecutionException e) {
            LOG.warn(e);
            wrapper.crashed(e);
        } finally {
            return lookupItems;
        }
    }

    /**
     * Creates a LookupElement given a CompletionItem
     *
     * @param item The CompletionItem
     * @return The corresponding LookupElement
     */
    @SuppressWarnings("WeakerAccess")
    public LookupElement createLookupItem(CompletionItem item) {
        Command command = item.getCommand();
        String detail = item.getDetail();
        String insertText = item.getInsertText();
        CompletionItemKind kind = item.getKind();
        String label = item.getLabel();
        Either<TextEdit, InsertReplaceEdit> textEditEither = item.getTextEdit();
        TextEdit textEdit = (textEditEither != null) ? textEditEither.getLeft() : null;
        InsertReplaceEdit insertReplaceEdit = (textEditEither != null) ? textEditEither.getRight() : null;
        List<TextEdit> addTextEdits = item.getAdditionalTextEdits();
        String presentableText = StringUtils.isNotEmpty(label) ? label : (insertText != null) ? insertText : "";
        String tailText = (detail != null) ? detail : "";
        LSPIconProvider iconProvider = GUIUtils.getIconProviderFor(wrapper.getServerDefinition());
        Icon icon = iconProvider.getCompletionIcon(kind);
        LookupElementBuilder lookupElementBuilder;

        String lookupString = null;
        if (textEdit != null) {
            lookupString = textEdit.getNewText();
        } else if (insertReplaceEdit != null) {
            lookupString = insertReplaceEdit.getNewText();
        } else if (StringUtils.isNotEmpty(insertText)) {
            lookupString = insertText;
        } else if (StringUtils.isNotEmpty(label)) {
            lookupString = label;
        }
        if (StringUtils.isEmpty(lookupString)) {
            return null;
        }
        // Fixes IDEA internal assertion failure in windows.
        lookupString = lookupString.replace(DocumentUtils.WIN_SEPARATOR, DocumentUtils.LINUX_SEPARATOR);

        lookupElementBuilder = LookupElementBuilder.create(getLookupStringWithoutPlaceholders(item, lookupString));

        lookupElementBuilder = addCompletionInsertHandlers(item, lookupElementBuilder, lookupString);

        if (kind == CompletionItemKind.Keyword) {
            lookupElementBuilder = lookupElementBuilder.withBoldness(true);
        }

        return lookupElementBuilder.withPresentableText(presentableText).withTypeText(tailText, true).withIcon(icon)
                .withAutoCompletionPolicy(AutoCompletionPolicy.SETTINGS_DEPENDENT);
    }

    private String getLookupStringWithoutPlaceholders(CompletionItem item, String lookupString) {
        if (item.getInsertTextFormat() == InsertTextFormat.Snippet) {
            return convertPlaceHolders(lookupString);
        } else {
            return lookupString;
        }
    }

    @SuppressWarnings("WeakerAccess")
    public LookupElementBuilder addCompletionInsertHandlers(CompletionItem item, LookupElementBuilder builder, String lookupString) {

        String label = item.getLabel();
        Command command = item.getCommand();
        List<TextEdit> addTextEdits = item.getAdditionalTextEdits();
        InsertTextFormat format = item.getInsertTextFormat();

        if (addTextEdits != null) {
            builder = builder.withInsertHandler((InsertionContext context, LookupElement lookupElement) -> invokeLater(() -> {
                if (format == InsertTextFormat.Snippet) {
                    context.commitDocument();
                    prepareAndRunSnippet(lookupString);
                }

                context.commitDocument();
                applyEdit(Integer.MAX_VALUE, toEither(addTextEdits), "Completion : " + label, false);
                if (command != null) {
                    executeCommands(Collections.singletonList(command));
                }
            }));
        } else if (command != null) {
            builder = builder.withInsertHandler((InsertionContext context, LookupElement lookupElement) -> {
                if (format == InsertTextFormat.Snippet) {
                    context.commitDocument();
                    prepareAndRunSnippet(lookupString);
                }
                context.commitDocument();
                executeCommands(Collections.singletonList(command));
            });
        } else {
            builder = builder.withInsertHandler((InsertionContext context, LookupElement lookupElement) -> {
                if (format == InsertTextFormat.Snippet) {
                    context.commitDocument();
                    prepareAndRunSnippet(lookupString);
                }
            });
        }

        return builder;
    }

    private int getCompletionPrefixLength(int offset) {
        return getCompletionPrefix(this.editor, offset).length();
    }

    private static final List<String> WHITESPACE_DELIMITERS = Arrays.asList(" \t\n\r".split(""));
    private static final List<String> SYMBOL_DELIMITERS = Arrays.asList("()[]{};,".split(""));

    @NotNull
    public String getCompletionPrefix(Editor editor, int offset) {
        StringBuilder s = new StringBuilder();
        String documentText = editor.getDocument().getText();
        for (int i = 0; i < offset; i++) {
            char singleLetter = documentText.charAt(offset - i - 1);
            val letterString = String.valueOf(singleLetter);
            if (WHITESPACE_DELIMITERS.contains(letterString) || SYMBOL_DELIMITERS.contains(letterString) || completionTriggers.contains(letterString)) {
                break;
            }
            s.append(singleLetter);
        }
        return s.reverse().toString();
    }

    @SuppressWarnings("WeakerAccess")
    public void prepareAndRunSnippet(String insertText) {

        List<SnippetVariable> variables = new ArrayList<>();
        // Extracts variables using placeholder REGEX pattern.
        Matcher varMatcher = Pattern.compile(SNIPPET_PLACEHOLDER_REGEX).matcher(insertText);
        while (varMatcher.find()) {
            variables.add(new SnippetVariable(varMatcher.group(), varMatcher.start(), varMatcher.end()));
        }

        variables.sort(Comparator.comparingInt(o -> o.startIndex));
        final String[] finalInsertText = {insertText};
        variables.forEach(var -> finalInsertText[0] = finalInsertText[0].replace(var.lspSnippetText, "$"));

        String[] splitInsertText = finalInsertText[0].split("\\$");
        finalInsertText[0] = String.join("", splitInsertText);

        TemplateImpl template = (TemplateImpl) TemplateManager.getInstance(getProject()).createTemplate(finalInsertText[0],
                "zigbrains-lsp");
        template.parseSegments();

        // prevent "smart" indent of next line...
        template.setToIndent(false);

        final int[] varIndex = {0};
        variables.forEach(var -> {
            template.addTextSegment(splitInsertText[varIndex[0]]);
            template.addVariable(varIndex[0] + "_" + var.variableValue, new TextExpression(var.variableValue),
                    new TextExpression(var.variableValue), true, false);
            varIndex[0]++;
        });
        // If the snippet text ends with a placeholder, there will be no string segment left to append after the last
        // variable.
        if (splitInsertText.length != variables.size()) {
            template.addTextSegment(splitInsertText[splitInsertText.length - 1]);
        }
        template.setInline(true);
        if (!variables.isEmpty()) {
            EditorModificationUtil.moveCaretRelatively(editor, -template.getTemplateText().length());
        }
        try {
            TemplateManager.getInstance(getProject()).startTemplate(editor, template);
        } catch (IllegalArgumentException e) {
            if (!e.getMessage().contains("Invalid offsets")) { //Discard exception, data race
                throw e;
            }
        }
    }

    private String convertPlaceHolders(String insertText) {
        return insertText.replaceAll(SNIPPET_PLACEHOLDER_REGEX, "");
    }

    /**
     * Returns the logical position given a mouse event
     *
     * @param e The event
     * @return The position (or null if out of bounds)
     */
    private LogicalPosition getPos(EditorMouseEvent e) {
        Point mousePos = e.getMouseEvent().getPoint();
        LogicalPosition editorPos = editor.xyToLogicalPosition(mousePos);
        Document doc = e.getEditor().getDocument();
        int maxLines = doc.getLineCount();
        if (editorPos.line >= maxLines) {
            return null;
        } else {
            int minY = doc.getLineStartOffset(editorPos.line) - (editorPos.line > 0 ?
                    doc.getLineEndOffset(editorPos.line - 1) : 0);
            int maxY = doc.getLineEndOffset(editorPos.line) - (editorPos.line > 0 ?
                    doc.getLineEndOffset(editorPos.line - 1) : 0);
            return (editorPos.column > minY && editorPos.column < maxY) ? editorPos : null;
        }
    }

    boolean applyEdit(List<Either<TextEdit, InsertReplaceEdit>> edits, String name, boolean setCaret) {
        return applyEdit(Integer.MAX_VALUE, edits, name, setCaret);
    }

    /**
     * Applies the given edits to the document
     *
     * @param version    The version of the edits (will be discarded if older than current version)
     * @param edits      The edits to apply
     * @param name       The name of the edits (Rename, for example)
     * @return True if the edits were applied, false otherwise
     */
    boolean applyEdit(int version, List<Either<TextEdit, InsertReplaceEdit>> edits, String name, boolean setCaret) {
        Runnable runnable = getEditsRunnable(version, edits, name, setCaret);
        writeAction(() -> {
            if (runnable != null) {
                CommandProcessor.getInstance()
                        .executeCommand(project, runnable, name, "LSPPlugin", editor.getDocument());
            }
            postEdit();
//            if (closeAfter) {
//                PsiFile file = PsiDocumentManager.getInstance(project).getPsiFile(editor.getDocument());
//                if (file != null) {
//                    FileEditorManager.getInstance(project).closeFile(file.getVirtualFile());
//                }
//            }
        });
        return runnable != null;
    }

    protected void postEdit() {

    }

    /**
     * Returns a Runnable used to apply the given edits and save the document
     * Used by WorkspaceEditHandler (allows to revert a rename for example)
     *
     * @param version The edit version
     * @param edits   The edits
     * @param name    The name of the edit
     * @return The runnable
     */
    public Runnable getEditsRunnable(int version, List<Either<TextEdit, InsertReplaceEdit>> edits, String name, boolean setCaret) {
        if (version < this.documentEventManager.getDocumentVersion()) {
            LOG.warn(String.format("Edit version %d is older than current version %d", version, this.documentEventManager.getDocumentVersion()));
            return null;
        }
        if (edits == null) {
            LOG.warn("Received edits list is null.");
            return null;
        }
        if (editor.isDisposed()) {
            LOG.warn("Text edits couldn't be applied as the editor is already disposed.");
            return null;
        }
        Document document = editor.getDocument();
        if (!document.isWritable()) {
            LOG.warn("Document is not writable");
            return null;
        }

        return () -> {
            // Creates a sorted edit list based on the insertion position and the edits will be applied from the bottom
            // to the top of the document. Otherwise all the other edit ranges will be invalid after the very first edit,
            // since the document is changed.
            List<LSPTextEdit> lspEdits = new ArrayList<>();
            edits.forEach(edit -> {
                if(edit.isLeft()) {
                    String text = edit.getLeft().getNewText();
                    Range range = edit.getLeft().getRange();
                    if (range != null) {
                        int start = DocumentUtils.LSPPosToOffset(editor, range.getStart());
                        int end = DocumentUtils.LSPPosToOffset(editor, range.getEnd());
                        lspEdits.add(new LSPTextEdit(text, start, end));
                    }
                } else if(edit.isRight()) {
                    String text = edit.getRight().getNewText();
                    Range range = edit.getRight().getInsert();

                    if (range != null) {
                        int start = DocumentUtils.LSPPosToOffset(editor, range.getStart());
                        int end = DocumentUtils.LSPPosToOffset(editor, range.getEnd());
                        lspEdits.add(new LSPTextEdit(text, start, end));
                    } else if ((range = edit.getRight().getReplace()) != null) {
                        int start = DocumentUtils.LSPPosToOffset(editor, range.getStart());
                        int end = DocumentUtils.LSPPosToOffset(editor, range.getEnd());
                        lspEdits.add(new LSPTextEdit(text, start, end));
                    }
                }
            });

            // Sort according to the start offset, in descending order.
            Collections.sort(lspEdits);

            lspEdits.forEach(edit -> {
                String text = edit.getText();
                int start = edit.getStartOffset();
                int end = edit.getEndOffset();
                if (StringUtils.isEmpty(text)) {
                    document.deleteString(start, end);
                    if (setCaret) {
                        editor.getCaretModel().moveToOffset(start);
                    }
                } else {
                    text = text.replace(DocumentUtils.WIN_SEPARATOR, DocumentUtils.LINUX_SEPARATOR);
                    if (end >= 0) {
                        if (end - start <= 0) {
                            document.insertString(start, text);
                        } else {
                            document.replaceString(start, end, text);
                        }
                    } else if (start == 0) {
                        document.setText(text);
                    } else if (start > 0) {
                        document.insertString(start, text);
                    }
                    if (setCaret) {
                        editor.getCaretModel().moveToOffset(start + text.length());
                    }
                }
                saveDocument();
            });
        };
    }

    /**
     * Sends commands to execute to the server and applies the changes returned if the future returns a WorkspaceEdit
     *
     * @param commands The commands to execute
     */
    public void executeCommands(List<Command> commands) {
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            commands.stream().map(c -> {
                if (c == null)
                    return null;
                ExecuteCommandParams params = new ExecuteCommandParams();
                params.setArguments(c.getArguments());
                params.setCommand(c.getCommand());
                return wrapper.getRequestManager().executeCommand(params);
            }).filter(Objects::nonNull).forEach(f -> {
                try {
                    f.get(Timeout.getTimeout(Timeouts.EXECUTE_COMMAND), TimeUnit.MILLISECONDS);
                    wrapper.notifySuccess(Timeouts.EXECUTE_COMMAND);
                } catch (TimeoutException te) {
                    LOG.warn(te);
                    wrapper.notifyFailure(Timeouts.EXECUTE_COMMAND);
                } catch (JsonRpcException | ExecutionException | InterruptedException e) {
                    LOG.warn(e);
                    wrapper.crashed(e);
                }
            });
        });
    }

    private void saveDocument() {
        FileDocumentManager.getInstance().saveDocument(editor.getDocument());
    }

    /**
     * Adds all the listeners
     */
    public void registerListeners() {
        editor.getCaretModel().addCaretListener(caretListener);
        // Todo - Implement
        // editor.getSelectionModel.addSelectionListener(selectionListener)
    }

    /**
     * Removes all the listeners
     */
    public void removeListeners() {
        editor.getCaretModel().removeCaretListener(caretListener);
        // Todo - Implement
        // editor.getSelectionModel.removeSelectionListener(selectionListener)
    }

    /**
     * Notifies the server that the corresponding document has been closed
     */
    public void documentClosed() {
        pool(() -> {
            if (this.isOpen) {
                isOpen = false;

                documentEventManager.documentClosed();
                EditorEventManagerBase.unregisterManager(this);
            } else {
                LOG.warn("Editor " + identifier.getUri() + " was already closed");
            }
        });
    }

    public void documentOpened() {
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            if (isOpen) {
                LOG.warn("Editor " + editor + " was already open");
            } else {
                documentEventManager.documentOpened();

                isOpen = true;
            }
        });
    }

    public void documentChanged(DocumentEvent event) {
        if (editor.isDisposed()) {
            return;
        }
        if (event.getDocument() == editor.getDocument()) {
            documentEventManager.documentChanged(event);
        } else {
            LOG.error("Wrong document for the EditorEventManager");
        }
    }

    /**
     * Notifies the server that the corresponding document has been saved
     */
    public void documentSaved() {
        pool(() -> {
            if (!editor.isDisposed()) {
                DidSaveTextDocumentParams params = new DidSaveTextDocumentParams(identifier, editor.getDocument().getText());
                wrapper.getRequestManager().didSave(params);
            }
        });
    }

    /**
     * Indicates that the document will be saved
     */
    //TODO Manual
    public void willSave() {
        if (wrapper.isWillSaveWaitUntil() && !needSave) {
            willSaveWaitUntil();
        } else
            pool(() -> {
                if (!editor.isDisposed()) {
                    wrapper.getRequestManager().willSave(new WillSaveTextDocumentParams(identifier, TextDocumentSaveReason.Manual));
                }
            });
    }

    public List<InlayHint> inlayHint() {
        var range = ApplicationUtil.computableReadAction(() -> {
            var start = DocumentUtils.offsetToLSPPos(editor, 0);
            var end = DocumentUtils.offsetToLSPPos(editor, editor.getDocument().getTextLength());
            return new Range(start, end);
        });
        var promise = getRequestManager().inlayHint(new InlayHintParams(FileUtils.editorToLSPIdentifier(editor), range));
        if (promise == null) {
            return Collections.emptyList();
        }
        try {
            var res = promise.get(Timeout.getTimeout(Timeouts.INLAY_HINTS), TimeUnit.MILLISECONDS);
            wrapper.notifySuccess(Timeouts.INLAY_HINTS);
            if (res != null) {
                return res;
            }
        } catch (TimeoutException e) {
            LOG.warn(e);
            wrapper.notifyFailure(Timeouts.CODEACTION);
        } catch (InterruptedException | JsonRpcException | ExecutionException e) {
            LOG.warn(e);
            wrapper.crashed(e);
        }
        return Collections.emptyList();
    }

    /**
     * If the server supports willSaveWaitUntil, the LSPVetoer will check if  a save is needed
     * (needSave will basically alternate between true or false, so the document will always be saved)
     */
    private void willSaveWaitUntil() {
        if (wrapper.isWillSaveWaitUntil()) {
            pool(() -> {
                if (editor.isDisposed()) {
                    return;
                }
                WillSaveTextDocumentParams params = new WillSaveTextDocumentParams(identifier,
                        TextDocumentSaveReason.Manual);
                CompletableFuture<List<TextEdit>> future = wrapper.getRequestManager().willSaveWaitUntil(params);
                if (future != null) {
                    try {
                        List<TextEdit> edits = future.get(Timeout.getTimeout(Timeouts.WILLSAVE), TimeUnit.MILLISECONDS);
                        wrapper.notifySuccess(Timeouts.WILLSAVE);
                        if (edits != null) {
                            invokeLater(() -> applyEdit(toEither(edits), "WaitUntil edits", false));
                        }
                    } catch (TimeoutException e) {
                        LOG.warn(e);
                        wrapper.notifyFailure(Timeouts.WILLSAVE);
                    } catch (JsonRpcException | ExecutionException | InterruptedException e) {
                        LOG.warn(e);
                        wrapper.crashed(e);
                    } finally {
                        needSave = true;
                        saveDocument();
                    }
                } else {
                    needSave = true;
                    saveDocument();
                }
            });
        } else {
            LOG.error("Server doesn't support WillSaveWaitUntil");
            needSave = true;
            saveDocument();
        }
    }
    // Tries to go to definition
    public boolean gotoDefinition(PsiElement element) {
        if (editor.isDisposed()) {
            return false;
        }
        val sourceOffset = element.getTextOffset();
        val loc = requestDefinition(DocumentUtils.offsetToLSPPos(editor, sourceOffset));
        if (loc == null)
            return false;

        return gotoLocation(project, loc);
    }

    // Tries to go to declaration / show usages based on the element which is
    public void gotoDeclarationOrUsages(PsiElement element) {
        if (editor.isDisposed()) {
            return;
        }
        val sourceOffset = element.getTextOffset();
        val loc = requestDeclaration(DocumentUtils.offsetToLSPPos(editor, sourceOffset));

        if (loc == null) {
            LSPReferencesAction referencesAction = (LSPReferencesAction) ActionManager.getInstance()
                                                                                      .getAction("ZBLSPFindUsages");
            if (referencesAction != null) {
                referencesAction.forManagerAndOffset(this, sourceOffset);
            }
            return;
        }

        String locUri = FileUtil.sanitizeURI(loc.getUri());

        if (identifier.getUri().equals(locUri)
                && sourceOffset >= DocumentUtils.LSPPosToOffset(editor, loc.getRange().getStart())
                && sourceOffset <= DocumentUtils.LSPPosToOffset(editor, loc.getRange().getEnd())) {
            LSPReferencesAction referencesAction = (LSPReferencesAction) ActionManager.getInstance()
                    .getAction("ZBLSPFindUsages");
            if (referencesAction != null) {
                referencesAction.forManagerAndOffset(this, sourceOffset);
            }
        } else {
            gotoLocation(project, loc);
        }
    }

    public static boolean gotoLocation(Project project, Location loc) {
        VirtualFile file = null;
        try {
            file = VfsUtil.findFileByURL(new URL(loc.getUri()));
        } catch (MalformedURLException e1) {
            LOG.warn("Syntax Exception occurred for uri: " + loc.getUri());
        }
        if (file != null) {
            OpenFileDescriptor descriptor = new OpenFileDescriptor(project, file);
            VirtualFile finalFile = file;
            invokeLater(() -> {
                FileEditorManager.getInstance(project).openTextEditor(descriptor, true);
                Editor srcEditor = FileUtils.editorFromVirtualFile(finalFile, project);
                if (srcEditor != null) {
                    Position start = loc.getRange().getStart();
                    LogicalPosition logicalPos = DocumentUtils.getTabsAwarePosition(srcEditor, start);
                    if (logicalPos != null) {
                        srcEditor.getCaretModel().moveToLogicalPosition(logicalPos);
                        srcEditor.getScrollingModel().scrollTo(logicalPos, ScrollType.CENTER);
                    }
                }
            });
            return true;
        } else {
            LOG.warn("Empty file for " + loc.getUri());
        }
        return false;
    }

    public void requestAndShowCodeActions() {
        invokeLater(() -> {
            if (editor.isDisposed()) {
                return;
            }
            if (annotations == null) {
                annotations = new ArrayList<>();
            }

            // sends code action request.
            int caretPos = editor.getCaretModel().getCurrentCaret().getOffset();
        pool(() -> {
            if (editor.isDisposed()) {
                return;
            }
            List<Either<Command, CodeAction>> codeActionResp = codeAction(caretPos);
            if (codeActionResp == null || codeActionResp.isEmpty()) {
                return;
            }
        invokeLater(() -> {
            if (editor.isDisposed()) {
                return;
            }
            codeActionResp.forEach(element -> {
                if (element == null) {
                    return;
                }
                if (element.isLeft()) {
                    Command command = element.getLeft();
                    annotations.forEach(annotation -> {
                        int start = annotation.getStartOffset();
                        int end = annotation.getEndOffset();
                        if (start <= caretPos && end >= caretPos) {
                            quickFixes.computeIfAbsent(annotation, (ignored) -> new ArrayList<>())
                                      .add(new QuickFixRequest(new LSPCommandFix(FileUtils.editorToURIString(editor), command),
                                                               new TextRange(start, end)));
                            codeActionSyncRequired = true;
                        }
                    });
                } else if (element.isRight()) {
                    CodeAction codeAction = element.getRight();
                    List<Diagnostic> diagnosticContext = codeAction.getDiagnostics();
                    annotations.forEach(annotation -> {
                        int start = annotation.getStartOffset();
                        int end = annotation.getEndOffset();
                        if (start <= caretPos && end >= caretPos) {
                            quickFixes.computeIfAbsent(annotation, (ignored) -> new ArrayList<>())
                                      .add(new QuickFixRequest(new LSPCodeActionFix(FileUtils.editorToURIString(editor),
                                                                                    codeAction), new TextRange(start, end)));
                            codeActionSyncRequired = true;
                        }
                    });

                    // If the code actions does not have a diagnostics context, creates an intention action for
                    // the current line.
                    if ((diagnosticContext == null || diagnosticContext.isEmpty()) && anonHolder != null && !codeActionSyncRequired) {
                        // Calculates text range of the current line.
                        int line = editor.getCaretModel().getCurrentCaret().getLogicalPosition().line;
                        int startOffset = editor.getDocument().getLineStartOffset(line);
                        int endOffset = editor.getDocument().getLineEndOffset(line);
                        TextRange range = new TextRange(startOffset, endOffset);

                        try {
                            this.anonHolder.newAnnotation(HighlightSeverity.INFORMATION, codeAction.getTitle())
                                           .range(range)
                                           .withFix(new LSPCodeActionFix(FileUtils.editorToURIString(editor), codeAction))
                                           .create();

                            @SuppressWarnings("unchecked")
                            SmartList<? extends Annotation> asList = (SmartList<Annotation>) this.anonHolder;
                            this.annotations.add(asList.get(asList.size() - 1));


                            diagnosticSyncRequired = true;
                        } catch (IllegalArgumentException e) {
                            //TODO Suppressed error, fix this somehow
                            e.printStackTrace();
                        }
                    }
                }
            });
            // If code actions are updated, forcefully triggers the inspection tool.
            if (codeActionSyncRequired) {
                // double-delay the update to ensure that the code analyzer finishes.
                invokeLater(this::updateErrorAnnotations);
            }
        });
        });
        });
    }

    /**
     * Triggers force full DaemonCodeAnalyzer execution.
     */
    private void updateErrorAnnotations() {
        computableReadAction(() -> {
            final PsiFile file = PsiDocumentManager.getInstance(project)
                    .getCachedPsiFile(editor.getDocument());
            if (file == null) {
                return null;
            }
            LOG.debug("Triggering force full DaemonCodeAnalyzer execution.");
            DaemonCodeAnalyzer.getInstance(project).restart(file);
            return null;
        });
    }

    public void initComplete() {
    }

    private static class LSPTextEdit implements Comparable<LSPTextEdit> {
        private String text;
        private int startOffset;
        private int endOffset;

        LSPTextEdit(String text, int start, int end) {
            this.text = text;
            this.startOffset = start;
            this.endOffset = end;
        }

        String getText() {
            return text;
        }

        int getStartOffset() {
            return startOffset;
        }

        int getEndOffset() {
            return endOffset;
        }

        @Override
        public int compareTo(@NotNull LSPTextEdit te) {
            return te.getStartOffset() - getStartOffset();
        }
    }

    static class SnippetVariable {
        String lspSnippetText;
        int startIndex;
        int endIndex;
        String variableValue;
        String intellijSnippetText;

        SnippetVariable(String text, int start, int end) {
            this.lspSnippetText = text;
            this.startIndex = start;
            this.endIndex = end;
            this.variableValue = getVariableValue(text);
        }

        private String getVariableValue(String lspVarSnippet) {
            if (lspVarSnippet.contains(":")) {
                return lspVarSnippet.substring(lspVarSnippet.indexOf(':') + 1, lspVarSnippet.lastIndexOf('}'));
            }
            return " ";
        }
    }
}
